/**
 * @file ppm_combinatorics_c.c
 * @brief combinatorial routines
 *
 * Copyright  (C)  2012  Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords: combinatorics
 * Maintainer: Thomas Jahns <jahns@dkrz.de>
 * URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <assert.h>
#include <inttypes.h>
#include <math.h>

#include "core/bittab.h"
#include "core/ppm_xfuncs.h"

int
PPM_prime_factorization_32(uint32_t n, uint32_t **factors)
{
  uint32_t lim, i, j, p;
  /* stores prime number candidates, where for each odd number i bit
   * i/2 is set if it's not a possible divisor, with the special case
   * that bit 0 encodes if 2 is not a divisor */
  PPM_Bittab *nprime_divs;
  int num_factors = 0;

  assert(n > 0);
  if (n == 1)
    return 0;
  lim = ceil(sqrt(n));
  p = n;
  nprime_divs = PPM_bittab_new(lim/2 + 1);
  if (!(p % 2))
  {
    do {
      p /= 2;
      ++num_factors;
    } while (!(p % 2));
  }
  else
    PPM_bittab_set_bit(nprime_divs, 0);
  i = 3;
  while (i <= lim)
  {
    if (!(p % i))
    {
      do {
        p /= i;
        ++num_factors;
      } while (!(p % i));
    }
    else
      PPM_bittab_set_bit(nprime_divs, i/2);
    for (j = 3*i; j <= lim; j += 2*i)
      PPM_bittab_set_bit(nprime_divs, j/2);
    i = PPM_bittab_get_next_clear_bitnum(nprime_divs, i/2);
    if (i == lim/2 + 1)
      break;
    i = 2 * i + 1;
  }
  /* assert(p == 1 || is_prime(p)); */
  if (p != 1)
    ++num_factors;
  if (factors)
  {
    unsigned long last_factor_pos = 0;
    i = 0;
    p = n;
    *factors = xmalloc(sizeof(**factors) * num_factors);
    if (!PPM_bittab_bit_is_set(nprime_divs, 0))
      while (!(p % 2))
      {
        (*factors)[i] = 2;
        p /= 2;
        ++i;
      }
    while (i < num_factors)
    {
      last_factor_pos
        = PPM_bittab_get_next_clear_bitnum(nprime_divs, last_factor_pos);
      if (last_factor_pos < lim/2 + 1)
      {
        uint32_t factor = last_factor_pos * 2 + 1;
        while (!(p % factor))
        {
          (*factors)[i] = factor;
          p /= factor;
          ++i;
        }
      }
      else
        break;
    }
    if (i < num_factors)
    {
      assert(p /= 1);
      (*factors)[i] = p;
    }
  }
  return num_factors;
}

/*
 * Local Variables:
 * license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
 * license-markup: "doxygen"
 * license-default: "bsd"
 * End:
 */
