/**
 * @file ppm_random_c.c
 * @brief C routines to use pseudo-random number generator in Fortran
 *
 * @copyright Copyright  (C)  2011  Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Thomas Jahns <jahns@dkrz.de>
 * URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <assert.h>
#include <limits.h>
#include <math.h>
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <cfortran.h>
#include "core.h"
#include "yarandom.h"
#include "ppm_random.h"
#include "ppm_extents.h"

int
PPM_irand(void)
{
  uint32_t i;
  while ((i = PPM_ya_random()) == UINT32_C(-2147483648))
    ;
  return (int)i;
}

FCALLSCFUN0(INT,PPM_irand,PPM_IRAND,ppm_irand)

int64_t
PPM_irand8(void)
{
  uint64_t i;
  while ((i = PPM_ya_random64()) == UINT64_C(-9223372036854775808))
    ;
  return (int64_t)i;
}

FCALLSCFUN0(INT64,PPM_irand8,PPM_IRAND8,ppm_irand8)

int
PPM_irandp()
{
  uint32_t i = PPM_ya_random() & ((UINT32_C(1)<<31) - UINT32_C(1));
  return (int)i;
}

FCALLSCFUN0(INT,PPM_irandp,PPM_IRANDP,ppm_irandp)

int
PPM_irandr(struct iinterval range)
{
  uint32_t i;
  uint32_t range_size = (uint32_t)range.last + (uint32_t)INT_MIN
    - (uint32_t)range.first + UINT32_C(1) + (uint32_t)INT_MIN;
  uint32_t range_max_multiple;
  assert(range.last >= range.first);
  range_max_multiple = (((uint32_t)-1) / range_size) * range_size;
  while ((i = PPM_ya_random()) > range_max_multiple)
    ;
  i = (i % range_size) + (uint32_t)range.first;
  return (int)i;
}

static inline int
PPM_irandr_f(struct iinterval *range)
{
  return PPM_irandr(*range);
}

FCALLSCFUN1(INT,PPM_irandr_f,PPM_IRANDR,ppm_irandr,PVOID)

void
PPM_irand_a(int *a, size_t n)
{
  size_t i;
  for (i = 0; i < n; ++i)
    a[i] = PPM_irand();
}

static inline void
PPM_irand_a_f(int *a, int n)
{
  assert(n >= 0);
  PPM_irand_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_irand_a_f,PPM_IRAND_A,ppm_irand_a,INTV,INT)
FCALLSCSUB2(PPM_irand_a_f,PPM_IRAND_A_2D,ppm_irand_a_2d,INTVV,INT)
FCALLSCSUB2(PPM_irand_a_f,PPM_IRAND_A_3D,ppm_irand_a_3d,INTVVV,INT)

void
PPM_irandp_a(int *a, size_t n)
{
  size_t i;
  for (i = 0; i < n; ++i)
    a[i] = PPM_irandp();
}

static inline void
PPM_irandp_a_f(int *a, int n)
{
  assert(n >= 0);
  PPM_irandp_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_irandp_a_f,PPM_IRANDP_A,ppm_irandp_a,INTV,INT)
FCALLSCSUB2(PPM_irandp_a_f,PPM_IRANDP_A_2D,ppm_irandp_a_2d,INTVV,INT)
FCALLSCSUB2(PPM_irandp_a_f,PPM_IRANDP_A_3D,ppm_irandp_a_3d,INTVVV,INT)

void
PPM_irandr_a(int *a, size_t n, struct iinterval range)
{
  size_t i;
  for (i = 0; i < n; ++i)
    a[i] = PPM_irandr(range);
}

static inline void
PPM_irandr_a_f(int *a, int n, struct iinterval *range)
{
  assert(n >= 0);
  PPM_irandr_a(a, (size_t)n, *range);
}

FCALLSCSUB3(PPM_irandr_a_f,PPM_IRANDR_A,ppm_irandr_a,INTV,INT,PVOID)
FCALLSCSUB3(PPM_irandr_a_f,PPM_IRANDR_A_2D,ppm_irandr_a_2d,INTVV,INT,PVOID)
FCALLSCSUB3(PPM_irandr_a_f,PPM_IRANDR_A_3D,ppm_irandr_a_3d,INTVVV,INT,PVOID)

double
PPM_drand()
{
  return PPM_ya_fsgrandom();
}

FCALLSCFUN0(DOUBLE,PPM_ya_fsgrandom,PPM_DRAND,ppm_drand)

double
PPM_drandp()
{
  return PPM_ya_frandom();
}

FCALLSCFUN0(DOUBLE,PPM_ya_frandom,PPM_DRANDP,ppm_drandp)

double
PPM_drandr(struct iinterval_dp range)
{
  double range_size, x;
  assert(range.last >= range.first);
  range_size = range.last - range.first;
  if (range.last == range.first)
    return range.first;
  x = range.first + PPM_ya_frandom() * nextafter(range_size, HUGE_VAL);
  return x;
}

static inline double
PPM_drandr_f(struct iinterval_dp *range)
{
  return PPM_drandr(*range);
}

FCALLSCFUN1(DOUBLE,PPM_drandr_f,PPM_DRANDR,ppm_drandr,PVOID)

void
PPM_drand_a(double *a, size_t n)
{
  size_t i;
  for (i = 0; i < n; ++i)
    a[i] = PPM_ya_fsgrandom();
}

static inline void
PPM_drand_a_f(double *a, int n)
{
  assert(n >= 0);
  PPM_drand_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_drand_a_f,PPM_DRAND_A,ppm_drand_a,DOUBLEV,INT)
FCALLSCSUB2(PPM_drand_a_f,PPM_DRAND_A_2D,ppm_drand_a_2d,DOUBLEVV,INT)
FCALLSCSUB2(PPM_drand_a_f,PPM_DRAND_A_3D,ppm_drand_a_3d,DOUBLEVVV,INT)

void
PPM_drandp_a(double *a, size_t n)
{
  size_t i;
  for (i = 0; i < n; ++i)
    a[i] = PPM_ya_frandom();
}

static inline void
PPM_drandp_a_f(double *a, int n)
{
  assert(n >= 0);
  PPM_drandp_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_drandp_a_f,PPM_DRANDP_A,ppm_drandp_a,DOUBLEV,INT)
FCALLSCSUB2(PPM_drandp_a_f,PPM_DRANDP_A_2D,ppm_drandp_a_2d,DOUBLEVV,INT)
FCALLSCSUB2(PPM_drandp_a_f,PPM_DRANDP_A_3D,ppm_drandp_a_3d,DOUBLEVVV,INT)

void
PPM_drandr_a(double *a, size_t n, struct iinterval_dp range)
{
  size_t i;
  for (i = 0; i < n; ++i)
    a[i] = PPM_drandr(range);
}

static inline void
PPM_drandr_a_f(double *a, int n, struct iinterval_dp *range)
{
  assert(n >= 0);
  PPM_drandr_a(a, (size_t)n, *range);
}

FCALLSCSUB3(PPM_drandr_a_f,PPM_DRANDR_A,ppm_drandr_a,DOUBLEV,INT,PVOID)
FCALLSCSUB3(PPM_drandr_a_f,PPM_DRANDR_A_2D,ppm_drandr_a_2d,DOUBLEVV,INT,PVOID)
FCALLSCSUB3(PPM_drandr_a_f,PPM_DRANDR_A_3D,ppm_drandr_a_3d,DOUBLEVVV,INT,PVOID)

float
PPM_frand()
{
  return PPM_ya_fsgrandomf();
}

FCALLSCFUN0(FLOAT,PPM_frand,PPM_FRAND,ppm_frand)

float
PPM_frandp()
{
  return PPM_ya_frandomf();
}

FCALLSCFUN0(FLOAT,PPM_frandp,PPM_FRANDP,ppm_frandp)

float
PPM_frandr(struct iinterval_sp range)
{
  float range_size, x;
  assert(range.last >= range.first);
  range_size = range.last - range.first;
  if (range.last == range.first)
    return range.first;
  x = range.first + PPM_ya_frandomf() * nextafterf(range_size, HUGE_VALF);
  return x;
}

static inline float
PPM_frandr_f(struct iinterval_sp *range)
{
  return PPM_frandr(*range);
}

FCALLSCFUN1(DOUBLE,PPM_frandr_f,PPM_FRANDR,ppm_frandr,PVOID)

void
PPM_frand_a(float *a, size_t n)
{
  size_t i;
  for (i = 0; i < n; ++i)
    a[i] = PPM_ya_fsgrandomf();
}

static inline void
PPM_frand_a_f(float *a, int n)
{
  assert(n >= 0);
  PPM_frand_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_frand_a_f,PPM_FRAND_A,ppm_frand_a,FLOATV,INT)
FCALLSCSUB2(PPM_frand_a_f,PPM_FRAND_A_2D,ppm_frand_a_2d,FLOATVV,INT)
FCALLSCSUB2(PPM_frand_a_f,PPM_FRAND_A_3D,ppm_frand_a_3d,FLOATVVV,INT)

void
PPM_frandp_a(float *a, size_t n)
{
  size_t i;
  for (i = 0; i < n; ++i)
    a[i] = PPM_ya_frandomf();
}

static inline void
PPM_frandp_a_f(float *a, int n)
{
  assert(n >= 0);
  PPM_frandp_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_frandp_a_f,PPM_FRANDP_A,ppm_frandp_a,FLOATV,INT)
FCALLSCSUB2(PPM_frandp_a_f,PPM_FRANDP_A_2D,ppm_frandp_a_2d,FLOATVV,INT)
FCALLSCSUB2(PPM_frandp_a_f,PPM_FRANDP_A_3D,ppm_frandp_a_3d,FLOATVVV,INT)

void
PPM_frandr_a(float *a, size_t n, struct iinterval_sp range)
{
  size_t i;
  for (i = 0; i < n; ++i)
    a[i] = PPM_frandr(range);
}

static inline void
PPM_frandr_a_f(float *a, int n, struct iinterval_sp *range)
{
  assert(n >= 0);
  PPM_frandr_a(a, (size_t)n, *range);
}

FCALLSCSUB3(PPM_frandr_a_f,PPM_FRANDR_A,ppm_frandr_a,FLOATV,INT,PVOID)
FCALLSCSUB3(PPM_frandr_a_f,PPM_FRANDR_A_2D,ppm_frandr_a_2d,FLOATVV,INT,PVOID)
FCALLSCSUB3(PPM_frandr_a_f,PPM_FRANDR_A_3D,ppm_frandr_a_3d,FLOATVVV,INT,PVOID)

static void
initIRand_f(MPI_Fint *comm_f, int *random_seed)
{
  MPI_Comm comm_c = MPI_COMM_NULL;
#if defined(USE_MPI)
  int flag = 0;
#if defined (__xlC__) && defined (_AIX)
#pragma omp critical
#endif
  if (MPI_Initialized(&flag) == MPI_SUCCESS && flag)
    comm_c = MPI_Comm_f2c((MPI_Fint)*comm_f);
#else
  comm_c = *comm_f;
#endif
  *random_seed = PPM_ya_rand_init(comm_c, *random_seed);
}

FCALLSCSUB2(initIRand_f,PPM_INITIALIZE_IRAND,ppm_initialize_irand,PVOID,PINT)

/*
 * Local Variables:
 * license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
 * license-markup: "doxygen"
 * license-default: "bsd"
 * End:
 */
