/*
 * @file ppm_random_mt_c.c
 * @brief multi-threaded array-filling PRNG routines
 *
 * Copyright  (C)  2012  Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * Keywords: PRNG OpenMP
 * @author Thomas Jahns <jahns@dkrz.de>
 * Maintainer: Thomas Jahns <jahns@dkrz.de>
 * URL: https://www.dkrz.de/redmine/projects/show/scales-ppm
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include "core/ppm_random.h"
#include "cfortran.h"

/* FIXME: these are a bit ugly because OpenMP 2.5 only allows for
 * signed loop counters */

void
PPM_irand_mt_a(int *a, size_t n)
{
#if defined(_OPENMP) && _OPENMP <= 200505
  ssize_t i;
  if (n > SSIZE_MAX)
  {
    ssize_t n2 = SSIZE_MAX;
#pragma omp for
    for (i = 0; i < n2; ++i)
      a[i] = PPM_irand();
    n -= (size_t)SSIZE_MAX;
    a += SSIZE_MAX;
  }
#pragma omp for
  for (i = 0; i < (ssize_t)n; ++i)
    a[i] = PPM_irand();
#else
  /* OpenMP 3.0 unsigned loop counters supported */
  size_t i;
#pragma omp for
  for (i = 0; i < n; ++i)
    a[i] = PPM_irand();
#endif
}

void
PPM_irandp_mt_a(int *a, size_t n)
{
#if defined(_OPENMP) && _OPENMP <= 200505
  ssize_t i;
  if (n > SSIZE_MAX)
  {
    ssize_t n2 = SSIZE_MAX;
#pragma omp for
    for (i = 0; i < n2; ++i)
      a[i] = PPM_irandp();
    n -= (size_t)SSIZE_MAX;
    a += SSIZE_MAX;
  }
#pragma omp for
  for (i = 0; i < (ssize_t)n; ++i)
    a[i] = PPM_irandp();
#else
  /* OpenMP 3.0 unsigned loop counters supported */
  size_t i;
#pragma omp for
  for (i = 0; i < n; ++i)
    a[i] = PPM_irandp();
#endif
}

void
PPM_irandr_mt_a(int *a, size_t n, struct iinterval range)
{
#if defined(_OPENMP) && _OPENMP <= 200505
  ssize_t i;
  if (n > SSIZE_MAX)
  {
    ssize_t n2 = SSIZE_MAX;
#pragma omp for
    for (i = 0; i < n2; ++i)
      a[i] = PPM_irandr(range);
    n -= (size_t)SSIZE_MAX;
    a += SSIZE_MAX;
  }
#pragma omp for
  for (i = 0; i < (ssize_t)n; ++i)
    a[i] = PPM_irandr(range);
#else
  /* OpenMP 3.0 unsigned loop counters supported */
  size_t i;
#pragma omp for
  for (i = 0; i < n; ++i)
    a[i] = PPM_irandr(range);
#endif
}

void
PPM_drand_mt_a(double *a, size_t n)
{
#if defined(_OPENMP) && _OPENMP <= 200505
  ssize_t i;
  if (n > SSIZE_MAX)
  {
    ssize_t n2 = SSIZE_MAX;
#pragma omp for
    for (i = 0; i < n2; ++i)
      a[i] = PPM_ya_fsgrandom();
    n -= (size_t)SSIZE_MAX;
    a += SSIZE_MAX;
  }
#pragma omp for
  for (i = 0; i < (ssize_t)n; ++i)
    a[i] = PPM_ya_fsgrandom();
#else
  /* OpenMP 3.0 unsigned loop counters supported */
  size_t i;
#pragma omp for
  for (i = 0; i < n; ++i)
    a[i] = PPM_ya_fsgrandom();
#endif
}

void
PPM_drandp_mt_a(double *a, size_t n)
{
#if defined(_OPENMP) && _OPENMP <= 200505
  ssize_t i;
  if (n > SSIZE_MAX)
  {
    ssize_t n2 = SSIZE_MAX;
#pragma omp for
    for (i = 0; i < n2; ++i)
      a[i] = PPM_ya_frandom();
    n -= (size_t)SSIZE_MAX;
    a += SSIZE_MAX;
  }
#pragma omp for
  for (i = 0; i < (ssize_t)n; ++i)
    a[i] = PPM_ya_frandom();
#else
  /* OpenMP 3.0 unsigned loop counters supported */
  size_t i;
#pragma omp for
  for (i = 0; i < n; ++i)
    a[i] = PPM_ya_frandom();
#endif
}

void
PPM_drandr_mt_a(double *a, size_t n, struct iinterval_dp range)
{
#if defined(_OPENMP) && _OPENMP <= 200505
  ssize_t i;
  if (n > SSIZE_MAX)
  {
    ssize_t n2 = SSIZE_MAX;
#pragma omp for
    for (i = 0; i < n2; ++i)
      a[i] = PPM_drandr(range);
    n -= (size_t)SSIZE_MAX;
    a += SSIZE_MAX;
  }
#pragma omp for
  for (i = 0; i < (ssize_t)n; ++i)
    a[i] = PPM_drandr(range);
#else
  /* OpenMP 3.0 unsigned loop counters supported */
  size_t i;
#pragma omp for
  for (i = 0; i < n; ++i)
    a[i] = PPM_drandr(range);
#endif
}

void
PPM_frand_mt_a(float *a, size_t n)
{
#if defined(_OPENMP) && _OPENMP <= 200505
  ssize_t i;
  if (n > SSIZE_MAX)
  {
    ssize_t n2 = SSIZE_MAX;
#pragma omp for
    for (i = 0; i < n2; ++i)
      a[i] = PPM_ya_fsgrandomf();
    n -= (size_t)SSIZE_MAX;
    a += SSIZE_MAX;
  }
#pragma omp for
  for (i = 0; i < (ssize_t)n; ++i)
    a[i] = PPM_ya_fsgrandomf();
#else
  /* OpenMP 3.0 unsigned loop counters supported */
  size_t i;
#pragma omp for
  for (i = 0; i < n; ++i)
    a[i] = PPM_ya_fsgrandomf();
#endif
}

void
PPM_frandp_mt_a(float *a, size_t n)
{
#if defined(_OPENMP) && _OPENMP <= 200505
  ssize_t i;
  if (n > SSIZE_MAX)
  {
    ssize_t n2 = SSIZE_MAX;
#pragma omp for
    for (i = 0; i < n2; ++i)
      a[i] = PPM_ya_frandomf();
    n -= (size_t)SSIZE_MAX;
    a += SSIZE_MAX;
  }
#pragma omp for
  for (i = 0; i < (ssize_t)n; ++i)
    a[i] = PPM_ya_frandomf();
#else
  /* OpenMP 3.0 unsigned loop counters supported */
  size_t i;
#pragma omp for
  for (i = 0; i < n; ++i)
    a[i] = PPM_ya_frandomf();
#endif
}

void
PPM_frandr_mt_a(float *a, size_t n, struct iinterval_sp range)
{
#if defined(_OPENMP) && _OPENMP <= 200505
  ssize_t i;
  if (n > SSIZE_MAX)
  {
    ssize_t n2 = SSIZE_MAX;
#pragma omp for
    for (i = 0; i < n2; ++i)
      a[i] = PPM_frandr(range);
    n -= (size_t)SSIZE_MAX;
    a += SSIZE_MAX;
  }
#pragma omp for
  for (i = 0; i < (ssize_t)n; ++i)
    a[i] = PPM_frandr(range);
#else
  /* OpenMP 3.0 unsigned loop counters supported */
  size_t i;
#pragma omp for
  for (i = 0; i < n; ++i)
    a[i] = PPM_frandr(range);
#endif
}

static inline void
PPM_drand_mt_a_f(double *a, int n)
{
  assert(n >= 0);
  PPM_drand_mt_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_drand_mt_a,PPM_DRAND_MT_A,ppm_drand_mt_a,DOUBLEV,INT)
FCALLSCSUB2(PPM_drand_mt_a,PPM_DRAND_MT_A_2D,ppm_drand_mt_a_2d,DOUBLEVV,INT)
FCALLSCSUB2(PPM_drand_mt_a,PPM_DRAND_MT_A_3D,ppm_drand_mt_a_3d,DOUBLEVVV,INT)

static inline void
PPM_irand_mt_a_f(int *a, int n)
{
  assert(n >= 0);
  PPM_irand_mt_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_irand_mt_a_f,PPM_IRAND_MT_A,ppm_irand_mt_a,INTV,INT)
FCALLSCSUB2(PPM_irand_mt_a_f,PPM_IRAND_MT_A_2D,ppm_irand_mt_a_2d,INTVV,INT)
FCALLSCSUB2(PPM_irand_mt_a_f,PPM_IRAND_MT_A_3D,ppm_irand_mt_a_3d,INTVVV,INT)

static inline void
PPM_irandp_mt_a_f(int *a, int n)
{
  assert(n >= 0);
  PPM_irandp_mt_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_irandp_mt_a_f,PPM_IRANDP_MT_A,ppm_irandp_mt_a,INTV,INT)
FCALLSCSUB2(PPM_irandp_mt_a_f,PPM_IRANDP_MT_A_2D,ppm_irandp_mt_a_2d,INTVV,INT)
FCALLSCSUB2(PPM_irandp_mt_a_f,PPM_IRANDP_MT_A_3D,ppm_irandp_mt_a_3d,INTVVV,INT)

static inline void
PPM_irandr_mt_a_f(int *a, int n, struct iinterval *range)
{
  assert(n >= 0);
  PPM_irandr_mt_a(a, (size_t)n, *range);
}

FCALLSCSUB3(PPM_irandr_mt_a_f,PPM_IRANDR_MT_A,ppm_irandr_mt_a,INTV,INT,PVOID)
FCALLSCSUB3(PPM_irandr_mt_a_f,PPM_IRANDR_MT_A_2D,ppm_irandr_mt_a_2d,
            INTVV,INT,PVOID)
FCALLSCSUB3(PPM_irandr_mt_a_f,PPM_IRANDR_MT_A_3D,ppm_irandr_mt_a_3d,
            INTVVV,INT,PVOID)

static inline void
PPM_drandp_mt_a_f(double *a, int n)
{
  assert(n >= 0);
  PPM_drandp_mt_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_drandp_mt_a_f,PPM_DRANDP_MT_A,ppm_drandp_mt_a,DOUBLEV,INT)
FCALLSCSUB2(PPM_drandp_mt_a_f,PPM_DRANDP_MT_A_2D,ppm_drandp_mt_a_2d,
            DOUBLEVV,INT)
FCALLSCSUB2(PPM_drandp_mt_a_f,PPM_DRANDP_MT_A_3D,ppm_drandp_mt_a_3d,
            DOUBLEVVV,INT)

static inline void
PPM_drandr_mt_a_f(double *a, int n, struct iinterval_dp *range)
{
  assert(n >= 0);
  PPM_drandr_mt_a(a, (size_t)n, *range);
}

FCALLSCSUB3(PPM_drandr_mt_a_f,PPM_DRANDR_MT_A,ppm_drandr_mt_a,DOUBLEV,INT,PVOID)
FCALLSCSUB3(PPM_drandr_mt_a_f,PPM_DRANDR_MT_A_2D,ppm_drandr_mt_a_2d,
            DOUBLEVV,INT,PVOID)
FCALLSCSUB3(PPM_drandr_mt_a_f,PPM_DRANDR_MT_A_3D,ppm_drandr_mt_a_3d,
            DOUBLEVVV,INT,PVOID)

static inline void
PPM_frand_mt_a_f(float *a, int n)
{
  assert(n >= 0);
  PPM_frand_mt_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_frand_mt_a_f,PPM_FRAND_MT_A,ppm_frand_mt_a,FLOATV,INT)
FCALLSCSUB2(PPM_frand_mt_a_f,PPM_FRAND_MT_A_2D,ppm_frand_mt_a_2d,FLOATVV,INT)
FCALLSCSUB2(PPM_frand_mt_a_f,PPM_FRAND_MT_A_3D,ppm_frand_mt_a_3d,FLOATVVV,INT)

static inline void
PPM_frandp_mt_a_f(float *a, int n)
{
  assert(n >= 0);
  PPM_frandp_mt_a(a, (size_t)n);
}

FCALLSCSUB2(PPM_frandp_mt_a_f,PPM_FRANDP_MT_A,ppm_frandp_mt_a,FLOATV,INT)
FCALLSCSUB2(PPM_frandp_mt_a_f,PPM_FRANDP_MT_A_2D,ppm_frandp_mt_a_2d,FLOATVV,INT)
FCALLSCSUB2(PPM_frandp_mt_a_f,PPM_FRANDP_MT_A_3D,ppm_frandp_mt_a_3d,
            FLOATVVV,INT)

static inline void
PPM_frandr_mt_a_f(float *a, int n, struct iinterval_sp *range)
{
  assert(n >= 0);
  PPM_frandr_mt_a(a, n, *range);
}

FCALLSCSUB3(PPM_frandr_mt_a_f,PPM_FRANDR_MT_A,ppm_frandr_mt_a,FLOATV,INT,PVOID)
FCALLSCSUB3(PPM_frandr_mt_a_f,PPM_FRANDR_MT_A_2D,ppm_frandr_mt_a_2d,
            FLOATVV,INT,PVOID)
FCALLSCSUB3(PPM_frandr_mt_a_f,PPM_FRANDR_MT_A_3D,ppm_frandr_mt_a_3d,
            FLOATVVV,INT,PVOID)

/*
 * Local Variables:
 * license-project-url: "https://www.dkrz.de/redmine/projects/show/scales-ppm"
 * license-markup: "doxygen"
 * license-default: "bsd"
 * End:
 */
